from __future__ import print_function

import pytest

from .. import filesystem

def test_move_nondirectory(tempdir):
    tempdir.write('afile', 'some data', 'utf-8')
    with pytest.raises(OSError) as excinfo:
        filesystem.move_dir('afile', 'afile2')
    assert 'not a directory' in str(excinfo.value)

def test_move_nonfile(tempdir):
    tempdir.makedir('adir')
    with pytest.raises(OSError) as excinfo:
        filesystem.move_file('adir', 'adir2')
    assert 'not a file' in str(excinfo.value)


def test_makepath_file(tempdir):
    tempdir.write('afile', 'some data', 'utf-8')

    with pytest.raises(OSError) as excinfo:
        filesystem.make_path('afile')
    assert 'not a directory' in str(excinfo.value)

def test_paths_to_tree_bad():
    with pytest.raises(ValueError):
        filesystem.paths_to_tree(['hello/world', 'hello/world/afile'])



from __future__ import print_function

import logging
import os
import shutil
from itertools import chain

log = logging.getLogger(__name__)


# --------------------------------------------------------------------------
#  Canonical path functions
# --------------------------------------------------------------------------
def cjoin(path, *paths, **kwargs):
    """
    Canonical join.
    Ensures the following invariants hold:
        - all paths relative to ./ are given without the ./ specifier,
          unless they refer to `.`, in which case ./ would be returned
        - all direcory paths end in '/'

    Examples
    >>> cjoin("foo", "bar", "baz")
    'foo/bar/baz'
    >>> cjoin("foo", "bar", "baz", is_dir=True)
    'foo/bar/baz/'
    >>> cjoin('./foo')
    'foo'
    >>> cjoin('./foo', is_dir=True)
    'foo/'
    >>> cjoin('.', is_dir=True)
    './'
    >>> cjoin('./')
    './'


    Parameters
    ----------
    path
    paths
    kwargs:
        is_dir: boolean
            is this path a directory?
    Returns
    -------

    """
    joined = os.path.join(path, *paths)
    if kwargs.get('is_dir', False) and not path.endswith('/'):
        joined += '/'

    if joined != './' and joined.startswith('./'):
        joined = joined[2:]

    return joined


def is_file(path):
    """
    tests whether a canonical path refers to a file or a directory
    """
    return not path.endswith('/')

def name(path):
    """
    Return the name of the last element in the path. if the path ends in a slash,
    then the name of the directory before the slash is returned
    """
    if path.endswith('/'):
        path = path[:-1]
    return os.path.basename(path)


# --------------------------------------------------------------------------
#  Common
# --------------------------------------------------------------------------
def save_cwd(function):
    def func(*args, **kwargs):
        _pwd = os.getcwd()
        try:
            return function(*args, **kwargs)
        finally:
            os.chdir(_pwd)

    return func


# --------------------------------------------------------------------------
#  File related
# --------------------------------------------------------------------------

def move_file(src, dst):
    if not os.path.isfile(src):
        raise OSError("Source path is not a file: {}".format(src))
    shutil.move(src, dst)


# --------------------------------------------------------------------------
#  Directory related
# --------------------------------------------------------------------------

def move_dir(src, dst):
    """Moves the source directory INTO the destination"""
    if not os.path.isdir(src):
        raise OSError("Source path is not a directory: {}".format(src))
    shutil.move(src, dst)

def collect_terminal_empty_dirs(root, move_tuples):
    """
    Returns a list of all empty directories.    
       With a recursive argument it can be shown that a list of (full)
       paths to all empty directories is sufficient to reconstruct
       all recursively empty directories as well.


       Arguments
       ---------
       path: the path to collect from

       move_tuples: list of 2-tuple
        pairs of absolute (src, dst) paths

    """

    # --- build root directory state as it currently is
    disk_state = []
    norm_root = os.path.normpath(os.path.abspath(root))
    for base, dirs, files in os.walk(root):
        norm_base = os.path.normpath(os.path.abspath(base))

        for d in dirs:
            disk_state.append(os.path.join(norm_base, d) + os.path.sep)
        for f in files:
            disk_state.append(os.path.join(norm_base, f))

    disk_tree = paths_to_tree(disk_state)

    def traverse(tree, parts, mkdir=False):
        """traverses parts and returns a (parent, name) tuple
           where name is a key into the parent, identifying the
           last item in parts.

           if mkdir is true, intermediary directories are created,
           the final path component is not created.
        """
        item = tree

        parent = None
        name = None

        if mkdir:
            for s in parts:
                item[s] = item.get(s, {})
                item, parent = item[s], item
                name = s
            del parent[name]
        else:
            for s in parts:
                item, parent = item[s], item
                name = s

        return (parent, name)

    def move(source, dest, tree):
        src_parts = _path_parts(source)
        dst_parts = _path_parts(dest)

        src_parent, src_name = traverse(tree, src_parts)
        src_item = src_parent[src_name]
        del src_parent[src_name]

        dst_parent, dst_name = traverse(tree, dst_parts, mkdir=True)

        assert dst_name not in dst_parent
        # perform the actual move
        dst_parent[dst_name] = src_item


    for (src, dst) in move_tuples:
        move(src, dst, disk_tree)


    empties = set()

    def collect_empty(tree, parts):
        if tree is None:
            return False
        elif tree == {}:
            return True

        all_empty = True
        for name in tree:
            value = tree[name]
            parts.append(name)

            if collect_empty(value, parts):
                empties.add(os.path.join(*parts))
            else:
                all_empty = False

            parts.pop()

        return all_empty

    src_parent, src_name = traverse(disk_tree, _path_parts(norm_root))
    collect_empty(src_parent[src_name], [root])

    return empties


def remove_empty_dirs(path):
    """Recursively removes empty direcotries contained within path"""

    # apparently the next loop does not return... ever
    for root, dirs, files in os.walk(path, topdown=False):  # pragma: no cover
        log.debug("[clean_empty.at] %s", path)

        # don't remove the source directory!
        if os.path.normpath(root) == os.path.normpath(path):
            break

        if len(os.listdir(root)) == 0:
            os.rmdir(root)
            log.debug("rmdir %s", root)



def make_path(path):
    """Creates intermediary directories so that the path exists"""
    path = os.path.abspath(path)
    if os.path.exists(path) and not os.path.isdir(path):
        raise OSError("File {} exists, but is not a directory".format(path))
    log.debug("make_path: %s", path)
    if not os.path.exists(path):
        os.makedirs(path)

def _path_parts(path):
    path = os.path.normpath(path)
    parts = path.split(os.path.sep)
    # the only empty parts should be the start
    # or the end
    return (_ for _ in parts if _)

def paths_to_tree(paths):
    """
    Create a dictionary structure representing a filesystem tree from paths.
    Where `paths` is a list of file and directory paths. If a path ends in
    in `os.path.sep` it is considerd a directory, else a file.

    Examples
    ---------
    >>> paths_to_tree(['/hello/there/'])
    {'hello': {'there': {}}}
    >>> paths_to_tree(['/hello/there'])
    {'hello': {'there': None}}
    >>> paths_to_tree(['/'])
    {}
    >>> paths_to_tree([])
    {}
    >>> paths_to_tree(set())
    {}
    >>> paths_to_tree(['hello/cruel/world', 'hello/cruel/earth/africa'])
    {'hello': {'cruel': {'world': None, 'earth': {'africa': None}}}}
    """

    tree = {}

    for path in paths:
        parent = None
        subtree = tree
        p = None
        try:
            for p in _path_parts(path):
                subtree[p] = subtree.get(p, {})
                subtree, parent = subtree[p], subtree
        except AttributeError:
            msg = 'Path previously classified as file, was\
                   accessed as directory. Ensure that all directory\
                   paths end in a `%s`' % (os.path.sep,)
            raise ValueError(msg)

        # case where the path references a file
        if p and not path[-1] in {os.path.sep, '/'}:
            parent[p] = None

    return tree






#!/usr/bin/python
# -*- coding: utf-8 -*-
"""The main implementation for calling pysorter from the commandline"""

from __future__ import print_function

import argparse
import logging
import os
import sys

from .oranize import Organizer

from .rules import RulesFileClassifier

log = logging.getLogger(__name__)

_last_sorter = None  # variable used during testing, set by main


def package_directory():
    """Return the absolute path to where pysorter is located"""
    return os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__))))

def validate_arguments(args):
    """
        Checks whether the paths and options
        provided as commandline arguments are valid.
        The application exits if they are not.
    """
    if args.filetypes:
        args.filetypes = os.path.abspath(args.filetypes)
    else:
        args.filetypes = os.path.join(package_directory(), 'filetypes.py')

    if not os.path.isfile(args.filetypes):
        raise OSError("Filetypes is not a file or does not exist: {}".format(args.filetypes))

    if args.unhandled_file:
        args.unhandled_file = os.path.abspath(args.unhandled_file)

    return args


def parse_args(args=None):
    """Create an argument parser"""
    from . import __version__
    parser = argparse.ArgumentParser(description='Reorganizes files and directories according to certain rules')

    parser.add_argument('directory',
                        help='The directory to be organized')

    parser.add_argument('-d', '--destination',
                        help='The destination directory to move organized files to.',
                        dest='dest_dir',
                        default=None)

    parser.add_argument('-p', '--process-dirs',
                        help='Should directories also be matched against the rules?',
                        action='store_true',
                        dest='do_process_dirs')

    parser.add_argument('-t', '--filetypes',
                        help='File containing regex rules [Default: filetypes.py]',
                        default=None)

    parser.add_argument('-u', '--unhandled-file',
                        help='Write the paths of all unhandled items to this file',
                        dest='unhandled_file')

    parser.add_argument('-r', '--recursive',
                        help='Recursively organize directories',
                        action='store_true',
                        dest='do_recurse')

    parser.add_argument('-c', '--remove-empty-dirs',
                        help='Recursively removes all empty directories in the directory being organized.',
                        action='store_true',
                        dest='do_remove_empty_dirs')

    parser.add_argument('-n', '--dry-run',
                        help='Prints out the changes that would occur, without actually executing them.',
                        action='store_true',
                        dest='dry_run')

    parser.add_argument('-V', '--version',
                        action='version',
                        version=__version__,
                        help='Prints out the current version of pysorter')

    return validate_arguments(parser.parse_args(args))

def main(args=None):
    global _last_sorter

    logging.basicConfig()
    args = parse_args(args)

    rules = RulesFileClassifier.load_file(args.filetypes)

    topass = dict(vars(args))

    # --- remove processed arguments
    del topass['directory']
    del topass['filetypes']
    del topass['unhandled_file']

    sorter = Organizer(args.directory, rules, **topass)
    sorter.organize()

    _last_sorter = sorter

    # write out all the unknown file types
    if args.unhandled_file:
        with open(args.unhandled_file, 'a') as ufile:
            for i in sorter.unhandled_paths:
                ufile.write(i)
                ufile.write('\n')
    return 0

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    sys.exit(main())



from __future__ import print_function

import os

from . import helper
from .. import commandline


def test_sort_only_filetypes_arg(tempdir):
    filetypes = {
        '.*\\.pdf$': 'docs/'
    }

    to_sort = 'source/'

    to_make = ['file.pdf']

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py']
    commandline.main(args)

    # --- compare sorted
    expected = ['docs/',
                'docs/file.pdf']

    tempdir.compare(expected=expected, path=to_sort)


def test_recursive_sort(tempdir):
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '--filetypes', 'filetypes.py']
    commandline.main(args)

    # --- compare sorted
    expected = ['foo/',
                'nested/',
                'nested/even/',
                'nested/even/deeper/',
                'docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3']
    tempdir.compare(expected=expected, path=to_sort)


def test_recursive_sort_with_directory_processing(tempdir):
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '--filetypes', 'filetypes.py', '--process-dirs']
    commandline.main(args)

    # --- compare sorted
    expected = ['foo/',
                'nested/',
                'nested/even/',
                'nested/even/deeper/',
                'docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3']
    tempdir.compare(expected=expected, path=to_sort)


def test_clean_empty(tempdir):
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '-c', '--filetypes', 'filetypes.py']
    commandline.main(args)

    # --- compare sorted
    expected = ['docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3']
    tempdir.compare(expected=expected, path=to_sort)


def test_duplicate_recursive(tempdir):
    filetypes = {
        '\.mp4$': 'mp4_files/'
    }

    to_sort = 's'

    to_make = ['s/300.mp4', 's/movie/300.mp4', ]

    helper.initialize_dir(tempdir, filetypes, to_make)

    args = [to_sort, '-r', '--filetypes', 'filetypes.py']
    commandline.main(args)

    # --- compare sorted
    expected = [
        'movie/',
        'movie/300.mp4',
        'mp4_files/',
        'mp4_files/300.mp4'
    ]
    tempdir.compare(expected=expected, path=to_sort)


def test_process_dirs_no_rules(tempdir):
    filetypes = {}

    to_sort = 'files/'

    to_make = [('movies',
                ['1.mp4', 'deep/2.mp4', '3.mp4']),
               'emptydir/'
               ]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py', '--process-dirs']
    commandline.main(args)

    # --- compare sorted
    expected = [] + helper.build_path_tree(to_make)
    tempdir.compare(expected=expected, path=to_sort)


def test_print_version(capsys):
    try:
        commandline.main(['--version'])
        assert False, 'did not print out version.'
    except SystemExit:
        out, err = capsys.readouterr()
        from .. import __version__

        # Python 3.4 & 3.5 print out the version on  stdout instead
        # of stderr.
        try:
            assert err.strip() == __version__
        except AssertionError:
            assert out.strip() == __version__


def test_write_unknown_types_correct(tempdir):
    filetypes = {
        r'\.pdf$': 'docs/'
    }

    unknown = 'unknown.txt'
    to_sort = 'files/'

    u_files = ['movie.mp4', 'kerry.mp3', 'phantom.mp3', ('direct', ['a.pdf'])]
    to_make = ['thesis.pdf'] + u_files

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    # --- compare sorted
    args = [to_sort, '-u', unknown, '--filetypes', 'filetypes.py']
    commandline.main(args)

    data = set(_ for _ in tempdir.read(unknown, encoding="utf8").split('\n') if _)

    unhandled = {'movie.mp4', 'kerry.mp3', 'phantom.mp3', 'direct/a.pdf', 'direct/'}
    assert not data.difference(unhandled)


def test_absolute_path(tempdir):
    filetypes = {
        r'.*': tempdir.path + '/files/docs/'
    }

    to_sort = 'files/'

    to_make = ['movie.mp4', 'kerry.mp3', 'phantom.mp3', ('direct', ['a.pdf'])]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py', '--process-dirs']
    commandline.main(args)

    # --- compare sorted
    expected = helper.build_path_tree(to_make, 'docs/') + ['docs/']
    tempdir.compare(expected=expected, path=to_sort)


def test_dry_run_has_no_effect(tempdir):
    filetypes = {
        r'\.pdf$': 'docs/',
        r'\.doc$': 'word/',
    }

    src_dir = 'src/'
    dst_dir = 'dst/'

    to_make = ['story.pdf',
               'subdirectory/news.pdf',
               'an/empty/dir/',
               'another/',
               'dir/story.pdf']

    src_tree = helper.build_path_tree(to_make, src_dir)
    root_tree = src_tree + [dst_dir, 'filetypes.py', src_dir]

    helper.initialize_dir(tempdir, filetypes, src_tree)
    tempdir.makedir(dst_dir)

    args = [os.path.join(tempdir.path, src_dir), '-nrc', '-d', dst_dir, '-t', 'filetypes.py']

    commandline.main(args)
    tempdir.compare(expected=root_tree, path='.')

    # --- strip the common prefix from all the move (src, dst) pairs
    move_pairs = set()
    common = len(tempdir.path) + 1  # for the seperator
    for (src, dst) in commandline._last_sorter.dry_mv_tuples:
        move_pairs.add((src[common:], dst[common:]))

    rmdir_paths = set()
    for path in commandline._last_sorter.dry_rmdir:
        rmdir_paths.add(path[common:])

    assert len(move_pairs) == 2
    assert ('src/story.pdf', 'dst/docs/story.pdf') in move_pairs
    assert ('src/subdirectory/news.pdf', 'dst/docs/news.pdf') in move_pairs

    assert 'src/subdirectory' in rmdir_paths
    assert 'src/an/empty/dir' in rmdir_paths
    assert 'src/an/empty' in rmdir_paths
    assert 'src/an' in rmdir_paths
    assert 'src/another' in rmdir_paths
    assert len(rmdir_paths) == 5


def test_dry_run_with_directory_move(tempdir):
    filetypes = {
        r'\.pdf$': 'docs/',
        r'\.doc$': 'word/',
        r'r/$': 'directories/'
    }

    src_dir = 'src/'
    dst_dir = 'dst/'

    to_make = ['story.pdf',
               'subdirectory/news.pdf',
               'an/empty/dir/',
               'another/',
               'stories/deep/books/book.pdf']

    src_tree = helper.build_path_tree(to_make, src_dir)
    root_tree = src_tree + [dst_dir, 'filetypes.py', src_dir]

    helper.initialize_dir(tempdir, filetypes, src_tree)
    tempdir.makedir(dst_dir)

    args = [os.path.join(tempdir.path, src_dir), '-nrcp', '-d', dst_dir, '-t', 'filetypes.py']

    commandline.main(args)
    tempdir.compare(expected=root_tree, path='.')

    # --- strip the common prefix from all the move (src, dst) pairs
    move_pairs = set()
    common = len(tempdir.path) + 1  # for the seperator
    for (src, dst) in commandline._last_sorter.dry_mv_tuples:
        move_pairs.add((src[common:], dst[common:]))

    rmdir_paths = set()
    for path in commandline._last_sorter.dry_rmdir:
        rmdir_paths.add(path[common:])

    assert len(move_pairs) == 5
    assert ('src/stories/deep/books/book.pdf', 'dst/docs/book.pdf') in move_pairs
    assert ('src/story.pdf', 'dst/docs/story.pdf') in move_pairs
    assert ('src/subdirectory/news.pdf', 'dst/docs/news.pdf') in move_pairs
    assert ('src/an/empty/dir/', 'dst/directories/dir') in move_pairs
    assert ('src/another/', 'dst/directories/another') in move_pairs

    should_be_removed = [
        'src/subdirectory',
        'src/an',
        'src/an/empty',
        'src/stories',
        'src/stories/deep',
        'src/stories/deep/books',
    ]

    assert set(should_be_removed) == rmdir_paths

def test_skip_recurse_on_file_ok(tempdir):
    tempdir.write('filetypes.py', "from pysorter.rules import SkipRecurse\nRULES = [('.*', SkipRecurse)]", 'utf-8')

    src_dir = 'src/'
    dst_dir = 'dst/'

    to_make = ['story.pdf']

    src_tree = helper.build_path_tree(to_make, src_dir)

    helper.initialize_dir(tempdir, None, src_tree)
    tempdir.makedir(dst_dir)

    args = [os.path.join(tempdir.path, src_dir), '-d', dst_dir, '-t', 'filetypes.py']

    commandline.main(args)

    root_tree = src_tree + [dst_dir, 'filetypes.py', src_dir]
    tempdir.compare(expected=root_tree, path='.')



from __future__ import print_function

import pytest

from . import helper
from .. import commandline


def test_bad_filetypes_file(tempdir):
    filetypes = {}

    to_sort = 'files/'

    to_make = []
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes']

    with pytest.raises(OSError):
        commandline.main(args)


def test_unknown_filetypes_created(tempdir):
    filetypes = {}

    to_sort = 'files/'

    to_make = ['']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-u', 'unknown_files', '-t', 'filetypes.py']

    expected = helper.build_path_tree([''], to_sort)
    expected += ['unknown_files', 'filetypes.py']

    commandline.main(args)
    tempdir.compare(expected=expected)


def test_crash_on_nonexistent_source(tempdir):
    with pytest.raises(OSError):
        args = ['tosort']
        commandline.main(args)


def test_file_as_directory(tempdir):
    with pytest.raises(OSError):
        tempdir.write('file', b'')
        args = ['file']
        commandline.main(args)



"""
Tests the three defined wildcard patterns in filetypes.py
"""

from . import helper
from .. import commandline
from ..filetypes import DIRECTORIES, FILES_WITHOUT_EXTENSION, FILES_WITH_EXTENSION


def test_directory_pattern(tempdir):
    filetypes = {
        DIRECTORIES[0]: DIRECTORIES[1]
    }

    to_sort = 'files/'

    to_make = ['foo/', 'bar/', 'nested/directory/', 'somefile']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs']

    # --- compare sorted
    to_make.remove('somefile')
    expected = helper.build_path_tree(to_make, 'directories/')
    expected += ['directories/', 'somefile']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_file_extension_pattern(tempdir):
    filetypes = {
        FILES_WITH_EXTENSION[0]: FILES_WITH_EXTENSION[1]
    }

    to_sort = 'files/'

    to_make = ['movie.mp4', 'story.doc', 'archive.tar.gz', 'noextension', ('adirectory', ['secret.txt']),
               ('directory.with.dots', [''])]
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs', '-r']

    # --- compare sorted
    expected = ['other/', 'other/mp4_files/', 'other/doc_files/', 'other/gz_files/',
                'other/mp4_files/movie.mp4', 'other/doc_files/story.doc', 'other/gz_files/archive.tar.gz',
                'other/txt_files/', 'other/txt_files/secret.txt',
                'noextension', 'adirectory/', 'directory.with.dots/']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_file_no_extension_pattern(tempdir):
    filetypes = {
        FILES_WITHOUT_EXTENSION[0]: FILES_WITHOUT_EXTENSION[1]
    }

    to_sort = 'files/'

    to_make = ['movie', 'story', 'archive', 'noextension', 'nested/file']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs', '-r']

    # --- compare sorted
    expected = ['other/', 'other/story', 'other/archive', 'other/movie',
                'other/noextension', 'nested/', 'other/file']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)



import os
import subprocess
from .. import commandline


def test_help_option():
    basedir = os.path.dirname(commandline.package_directory())
    os.chdir(basedir)

    try:
        env = {'PYTHONPATH': basedir}
        output = subprocess.check_output(
                    ['python', '-m', 'pysorter', '--help'],
                    stderr=subprocess.STDOUT,
                    env=env
        )
    except subprocess.CalledProcessError as e:
        output = e.output

    assert '-d DEST_DIR' in output.decode('utf-8')



"""
    Contains the core logic of the sorting implementation
"""
from __future__ import print_function

import logging

import os

from . import rules
from . import filesystem as fs

log = logging.getLogger(__name__)

def escape_singles(string):
    """escapes all single quotes in a string"""
    return string.replace("'", r"\'")


class Organizer(object):
    def __init__(self,
                 source_dir,
                 sort_rule,
                 no_process=None,
                 dest_dir=None,

                 dry_run=False,

                 do_process_dirs=False,
                 do_recurse=False,
                 do_remove_empty_dirs=False):
        """
        Construct a new instance of Organizer for organizing some directory
        using certain parameters

        Parameters
        -----------
        source_dir: string
           path to the directory that must be organized

        sort_rule: function(path: str) --> destination: {str, class}
            Function that takes the path of an entity (file / directory) 
            as only argument and returns the destination path where it should 
            be moved to.
            
            If `path` ends in `/` then the path is a directory,
            else it is a file.
            
            If the destination ends in `/` then the the entity will me moved *inside*
            of the destination, otherwise the entity will be moved to *destination*.
            
            The function may also return or raise rules.Unhandled, rules.Skip or
            rules.SkipRecurse.
            

        no_process: set()
            names of toplevel directories that should not be traversed

        dest_dir: string
            the directory to organize your files into, instead of doing it in-place,
            if not set, it will be set to `source_dir`

        dry_run: boolean
            prints all file and directory that would occur changes without
            actually performing organization.

        do_recurse: boolean
            should directories in the `source_dir` be traversed recursively?

        do_remove_empty_dirs: boolean
            toggles recursive empty directory removal
        """
        dest_dir = dest_dir or source_dir

        if not os.path.isdir(source_dir):
            raise OSError("Directory to organize does not exist or is a file: {}".format(source_dir))

        if not os.path.isdir(dest_dir):
            log.warn("Destination directory does not exist, creating: %s", dest_dir)
            fs.make_path(dest_dir)

        self.path_source = os.path.abspath(source_dir)
        self.path_dest = os.path.abspath(dest_dir)

        self.sort_rule = sort_rule

        self.unhandled_paths = set()

        # directories or files that should not be processed paths are relative
        # to the directory to be sorted and should exclude any starting ./

        self.no_process = no_process or set()
        self.no_recurse = set()

        self.is_dry_run = dry_run

        self.do_remove_empty_dirs = do_remove_empty_dirs
        self.do_recurse = do_recurse
        self.do_process_dirs = do_process_dirs

        # --- variables used in a dry run

        # functions as an overlay of the destination directory,
        # so that changes in the source directory may be reflected through it.

        self.dry_src = set()
        self.dry_dst = set()
        self.dry_mv_tuples = []
        self.dry_rmdir = []

        self.files = {}

    @fs.save_cwd
    def sortrule_destination(self, path):
        """
        Invokes self.sortrule.destination, ensured that  Skip, SkipReturn or Unhandled
        are raised when the function returns them.
        """
        retval = self.sort_rule(path)
        if retval in rules.actions:
            raise retval()
        return retval

    @fs.save_cwd
    def organize(self):
        """
        The `main` function for organization.
        """
        os.chdir(self.path_source)

        def filter_no_process(base, alist, skipset=self.no_process):
            """
            Removes items from a list returned by os.walk, in place.
            """
            L = len(alist)
            for idx, i in enumerate(list(reversed(alist))):
                i = fs.cjoin(base, i)
                if i in skipset:
                    del alist[L - idx - 1]

        for base, dirs, files in os.walk('.'):
            filter_no_process(base, files)
            filter_no_process(base, dirs)

            for file in files:
                self.process(fs.cjoin(base, file))

            if self.do_process_dirs:
                for dir in dirs:
                    self.process(fs.cjoin(base, dir, is_dir=True))
                filter_no_process(base, dirs, skipset=self.no_recurse)
                self.no_recurse = set()  # clear out

            if not self.do_recurse:
                del dirs[:]

        if self.do_remove_empty_dirs:
            if self.is_dry_run:
                self.dry_rmdir = fs.collect_terminal_empty_dirs(self.path_source, self.dry_mv_tuples)
                for path in self.dry_rmdir:
                    print("rmdir '{}'".format(escape_singles(path)))
            else:
                fs.remove_empty_dirs(self.path_source)


    def process(self, src):
        """
        Take a single path to a directory or a file and
        apply some action to it, as defined by the sorting rule
        """
        name = fs.name(src)

        try:
            raw_dst = self.sortrule_destination(src)

            # a relative destination
            if not os.path.isabs(raw_dst):
                raw_dst = os.path.join(self.path_dest, raw_dst)

            if raw_dst.endswith('/'):
                # the destination is a directory, so the
                # file / directory must go INSIDE
                dst = fs.cjoin(raw_dst, name)
            else:
                # the destination is absolute, file must go there
                dst = fs.cjoin(raw_dst)

        except rules.Unhandled:
            self.unhandled_paths.add(src)
            return
        except rules.Skip:
            return
        except rules.SkipRecurse:
            if src.endswith('/'):
                # strip the trailing slash because the path will be compared
                # with `dirs` returned by os.walk (which do not contain
                # a trailing slash `/`)
                self.no_recurse.add(src.rstrip('/'))
            log.warning('SkipRecurse cannot be used with a file argument, Skip assumed: %s', src)
            return

        if os.path.exists(dst) or (dst in self.dry_dst):
            log.info("destination exists: `%s` --> `%s`", src, dst)
            return

        if self.is_dry_run:
            abs_src = os.path.join(self.path_source, src)
            self.dry_src.add(abs_src)
            self.dry_dst.add(dst)

            self.dry_mv_tuples.append((abs_src, dst))

            if not fs.is_file(src):
                self.no_recurse.add(src.rstrip('/'))
            print("mv '{}' '{}'".format(escape_singles(abs_src), escape_singles(dst)))
            return

        fs.make_path(os.path.dirname(dst))
        log.info("move {} --> {}".format(src, dst))
        if fs.is_file(src):
            fs.move_file(src, dst)
        else:
            fs.move_dir(src, dst)



from __future__ import print_function

import pytest

from . import helper
from .. import rules
from .. import commandline


def test_keyword_captures(tempdir):
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{keyword}/{name}.{ext}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'cruel/',
        'cruel/hello.pdf']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_numerical_captures(tempdir):
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['file.pdf',
               'awesome_song.mp3']

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py']
    commandline.main(args)

    # --- compare sorted
    expected = ['docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3']
    tempdir.compare(expected=expected, path=to_sort)


def test_bad_keyword_captures(tempdir):
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{unknown}/{keyword}/{name}.{ext}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    with pytest.raises(ValueError):
        commandline.main(args)


def test_bad_numerical_capture(tempdir):
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{keyword}/{name}.{ext}/{4}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    with pytest.raises(ValueError):
        commandline.main(args)


def test_callable_as_action(tempdir):
    def comparison_function(match, entity):
        return 'foobar/'

    filetypes = {
        '\.pdf$': comparison_function
    }

    to_sort = 'files/'
    dest = 'sorted/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '-d', dest]

    commandline.main(args)
    # --- compare sorted
    expected = [
        'foobar/',
        'foobar/hello_cruel.pdf']
    commandline.main(args)
    tempdir.compare(expected=expected, path=dest)


def test_rule_behaviour_into_directory_for_file(tempdir):
    filetypes = {
        '\.pdf$': 'pdf/'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/hello_cruel.pdf']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_rule_behaviour_to_for_file(tempdir):
    filetypes = {
        '.*\.pdf$': 'pdf'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'pdf']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_rule_behaviour_into_directory_for_dir(tempdir):
    filetypes = {
        'foo/$': 'pdf/'
    }

    to_sort = 'files/'

    to_make = ['foo/']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs']

    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/foo/']
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_rule_behaviour_to_for_dir(tempdir):
    filetypes = {
        'foo/$': 'pdf'
    }

    to_sort = 'files/'

    to_make = ['foo/']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs']

    # --- compare sorted
    expected = [
        'pdf/'
    ]
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_skip_pattern_file(tempdir):
    filetypes = {
        '\.xml$': rules.Skip,
        '\.pdf$': 'pdf/'
    }

    to_sort = 'files/'
    to_make = ['hello_cruel.pdf', 'config.xml']
    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']
    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/hello_cruel.pdf',
        'config.xml'
    ]
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)


def test_skip_recurse_directory(tempdir):
    filetypes = {
        'foo/$': rules.SkipRecurse,
        '\.pdf$': 'pdf/'
    }

    to_sort = 'files/'
    to_make = [('foo', ['docfoo.pdf']), ('bar', ['docbar.pdf'])]

    helper.initialize_dir(tempdir, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '-p', '-t', 'filetypes.py']
    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/docbar.pdf',
        'bar/',
        'foo/',
        'foo/docfoo.pdf'
    ]
    commandline.main(args)
    tempdir.compare(expected=expected, path=to_sort)




from __future__ import print_function

import pytest

from .. import rules

def test_no_rules_in_file(tempdir):
    tempdir.write('filetypes.py', 'import os', 'utf-8')
    with pytest.raises(RuntimeError) as excinfo:
        rules.RulesFileClassifier.load_file('filetypes.py')
    assert 'missing RULES' in str(excinfo.value)

def test_bad_rule(tempdir):
    tempdir.write('filetypes.py', "RULES = [('a', 1)]", 'utf-8')
    with pytest.raises(ValueError) as excinfo:
        rules.RulesFileClassifier.load_file('filetypes.py')
    assert 'unhandled type in rule list' in str(excinfo.value).lower()



"""
Module for the base regex rule implementation
"""
from __future__ import print_function

import logging
import re

log = logging.getLogger(__name__)

# --------------------------------------------------------------------------
# Actions
# --------------------------------------------------------------------------
class Unhandled(Exception):
    """
    Raised or returned when a path cannot be handled by the
    function.
    """

class Skip(Exception):
    """
    Riased or returned when a path should be skipped from moving.
    """

class SkipRecurse(Exception):
    """
    Raised or returned when a directory path should not
    be recursed into. 
    
    The directory itself may still be relocated though.
    """

actions = frozenset([Unhandled, Skip, SkipRecurse])


class RulesFileClassifier(object):
    """
    Default rule implementation that 
    uses the regex definitions as given in a `filetypes.py` specification file.
    
    """

    def __init__(self, rules):
        super(RulesFileClassifier, self).__init__()
        self.rules = rules

    def first_match(self, finditer):
        for match in finditer:
            return match
        return None

    def destination(self, path):
        for R, function in self.rules:
            match = self.first_match(R.finditer(path))
            if match:
                return function(match, path)
        raise Unhandled

    @classmethod
    def load_file(cls, path):
        """
        Loads sorting rules from a text file and return 
        a RulesFileClassifier containing all the sorting entries.
        """
        namespace = {'__builtins__': __builtins__}
        with open(path, 'r') as f:
            exec(f.read(), namespace)

        if 'RULES' not in namespace:
            msg = "Configuration file missing RULES: {}".format(path)
            raise RuntimeError(msg)

        rules = []
        for regex, destination in namespace['RULES']:
            pattern = re.compile(regex)
            matcher = None

            if is_string(destination):
                # destination --> format string
                matcher = make_regex_rule_function(pattern, destination)
            elif destination in actions:
                # constant action ex. Skip
                matcher = make_constant_function(destination)
            elif callable(destination):
                # custom processing_function(re_match, filepath)
                matcher = destination
            else:
                msg = "Unhandled type in rule list. " \
                      "Second item in pair must be " \
                      "callable, string or action: " + repr(destination)
                raise ValueError(msg)

            rules.append((pattern, matcher))

        return cls(rules)

    def __call__(self, path):
        return self.destination(path)


def is_string(obj):
    import sys

    PY3 = sys.version_info[0] == 3

    if PY3:
        string_types = str,  # pragma: no cover
    else:
        string_types = basestring, str  # pragma: no cover
    return isinstance(obj, string_types)


def make_constant_function(action):
    """Return a function that always returns `action`, regardless of its arguments"""

    def function(*args):
        return action

    return function


def make_regex_rule_function(pattern, dstfmt):
    """
    Return a path processing function. 
    """

    def process(re_match, path):
        """
        
        Parameters
        ----------
        re_match: regex match
        path: str
            original path that was used to create `match`

        Returns
        -------

        """
        # FIXME only re_match against the filename
        try:
            pargs = [re_match.group(0)] + list(re_match.groups())
            destination = dstfmt.format(*pargs,
                                        **re_match.groupdict())
        except IndexError:
            raise ValueError("Destination string placeholders out of range: {}".format(dstfmt))
        except KeyError as e:
            msg = "Destination string placeholder " \
                  "unknown key {}: {}".format(repr(e.args[0]), dstfmt)
            raise ValueError(msg)
        log.debug("process: RE[%s](%s) --> %s", pattern.pattern, path, destination)
        return destination

    return process
